{ COLUMNS.PAS : Text column layout and page/line-number filter

  Title   : COLUMNS
  Language: Turbo Pascal v7.0
  Version : 2.3
  Date    : Feb 13, 2000
  Author  : J.R. Ferguson, Amsterdam, The Netherlands
  Usage   : Refer procedure Help
  Download: http://hello.to/ferguson
            To compile this sourcefile, you will need some units from the Pascal
            library JRFPAS that can be found at the same Internet site.
  E-mail  : j.r.ferguson@iname.com
}

{$V-}
{$R+}

program Columns;
uses DefLib, ArgLib, ChrLib, StpLib;


const
  OptChr    = '/';
  MaxLine   = 255;
  MaxLPP    = 100;
  MaxCols   = MaxLine;
  MaxWid    = MaxLine;
  MaxSep    = MaxLine;
  MaxTitle  = MaxLine;
  MaxFname  = 79;
  IObufSiz  = 16 * 1024;

  { Error codes and messages: }
  RcOk      = 0;
  ErrArg    = 1;
  ErrInp    = 2;
  ErrOut    = 3;

  ErrMsg    : array[ErrInp..ErrOut] of StpTyp =
 ('File not found : ',
  'Can''t open output : '
 );


type
  LineStp   = String[MaxLine];
  LineInd   = 0..MaxLine;
  PageTyp   = array[1..MaxLPP] of LineStp;
  PageInd   = 0..MaxLPP;

var
  InpFname,
  OutFname  : StpTyp;
  InpFvar,
  OutFvar   : Text;
  InpBuf,
  OutBuf    : array[1..IObufSiz] of char;
  InpOpen,
  OutOpen   : boolean;
  rc        : integer;

  Cols      : LineInd; { columns per page }
  PagLen    : PageInd; { lines per page }
  PagWid    : LineInd; { characters per line }
  Sep       : StpTyp;  { column separator }
  Title     : StpTyp;  { page title }
  OptPagNum : boolean; { page numbering switch }
  OptLinNum : boolean; { line numbering switch }
  OptDblPrn : boolean; { doublesided printing }

  ColWid    : LineInd; { characters per column }
  SepWid    : LineInd; { separator width }
  PageNum   : integer; { current page number }
  LineNum   : longint; { current line number }
  LnumWid   : LineInd; { width of line number field }
  Page      : PageTyp; { Current Page }
  NLines    : PageInd; { Number of lines in current page }
  CurLine   : LineStp; { Current line }


{
--- Command line parsing routines ---
}


procedure Help;
procedure wl(s: StpTyp); begin writeln(s); end;
begin
wl('COLUMNS v2.3 Text column layout and page/line-number filter. JR Ferguson');
wl('');
wl('Usage : COLUMNS [<in] [>out] [/option[...] [...]]');
wl('   or : COLUMNS  in [[>]out] [/option[...] [...]]');
wl('');
wl('Options:                                                var range  default');
wl('  Ln    page length n lines                              n  1-100   60');
wl('  Wn    line width n chars                               n  1-255   80');
wl('  Cn    n columns per page                               n  1-255   1');
wl('  Ss    column separator s                               s          " "');
wl('  Ts    page title s                                     s          ""');
wl('  N[n]  number lines,n digits (adds n+2 to line width)   n  0-255   4');
wl('  P     number pages (adds 2 to page length)');
wl('  D     doublesided printing (even nr of pages, symmetrical header)');
wl('  ?     display this help text (output may be redirected)');
wl('');
wl('Remarks:');
wl('- If a parameter contains spaces, put the whole argument in quotes, as in:');
wl('    COLUMNS README.DOC >PRN /P "/T=== README.DOC ==="');
wl('- This filter may be used in a pipeline command.');
end;


function ReadUns(var arg: StpTyp): integer;
var
  n: integer;
  c: char;
begin
  n:= 0;
  c:= StpcGet(arg);
  while IsDigit(c) do begin
    n:= 10*n + ord(c) - 48;
    c:= StpcGet(arg);
  end;
  if c<>chr(0) then StpcIns(arg,c,1);
  ReadUns:= n
end;


function NeedUns(var arg: StpTyp; low,high: integer): integer;
var n: integer;
begin
  n:= 0;
  if not IsDigit(StpcRet(arg,1)) then rc:= ErrArg
  else begin
    n:= ReadUns(arg);
    if (n < low) or (n > high) then rc:= ErrArg;
  end;
  NeedUns:= n;
end;


function ReadStp(var arg: StpTyp): StpTyp;
var s: StpTyp;
begin
  StpCpy(s,arg);
  StpDel(arg,1,StpLen(s));
  ReadStp:= s;
end;


procedure ReadOpt(arg: StpTyp);
begin
  StpDel(arg,1,1);
  while (rc=RcOk) and not StpEmpty(arg) do case ToUpper(StpcGet(arg)) of
    'C': Cols      := NeedUns(arg,1,MaxCols);
    'W': PagWid    := NeedUns(arg,1,MaxWid);
    'L': PagLen    := NeedUns(arg,1,MaxLPP);
    'S': Sep       := ReadStp(arg);
    'T': Title     := ReadStp(arg);
    'P': OptPagNum := true;
    'D': OptDblPrn := true;
    'N': begin
           OptLinNum := true;
           if IsDigit(StpcRet(arg,1)) then LnumWid:= ReadUns(arg);
         end;
    '/': {just skip}
    else rc:= ErrArg;
  end;
end;


procedure ReadArgs;
var i,f : ArgInd;
    arg : StpTyp;
begin
  StpCreate(InpFname); StpCreate(OutFname); { Standard input/output }
  GetArgs;
  i:= 0; f:= 0;
  while (i < ArgC) and (rc = RcOk) do begin
    Inc(i); StpCpy(arg,ArgV[i]);
    case StpcRet(arg,1) of
      '/' : ReadOpt(arg);
      else case f of
        0 : begin Inc(f); StpNCpy(InpFname,arg,MaxFname); end;
        1 : begin Inc(f); StpNCpy(OutFname,arg,MaxFname); end;
       else rc:= ErrArg;
      end;
    end;
  end;
end;


{
--- I/O routines ---
}


procedure OpenInp;
begin
  StpNCpy(InpFname,InpFname,MaxFname);
  {$I-} Assign(InpFvar,InpFname); SetTextBuf(InpFvar,InpBuf); reset(InpFvar);
  {$I+} if IOresult <> 0 then rc:= ErrInp
  else InpOpen:= true;
end;


procedure OpenOut;
begin
  StpNCpy(OutFname,OutFname,MaxFname);
  {$I-} Assign(OutFvar,OutFname); SetTextBuf(OutFvar,OutBuf); rewrite(OutFvar);
  {$I+} if IOresult <> 0 then rc:= ErrOut else OutOpen:= true;
end;


function GetLine: Boolean;
begin
  if eof(InpFvar) then GetLine:= false
  else begin
    ReadLn(InpFvar,CurLine);
    GetLine:= true;
  end;
end;


{
--- Main line ---
}


function GetPage: boolean;
var r    : PageInd;
    c    : LineInd;
    Line : LineStp;
    width: LineInd;
    ok   : boolean;
begin
  for r:= 1 to PagLen do Page[r]:= '';
  c:= 1; width:= 0; r:= 0; Nlines:= 0;
  ok:= GetLine;
  while ok do begin
    Inc(r); if c=1 then Inc(NLines);
    StpRTS(CurLine);
    if c > 1 then begin
      StpFill(Page[r],' ',width); StpCat(Page[r],Sep);
    end;
    StpNCat(Page[r],CurLine,ColWid);
    if r < PagLen then ok:= GetLine
    else begin
      if c < Cols then begin
        width:= c*ColWid + pred(c)*SepWid;
        inc(c); r:= 0;
        ok:= GetLine;
      end
      else ok:= false;
    end;
  end;
  GetPage:= Nlines > 0;
end;

procedure PutHeader;
var Header, PagNumStr: StpTyp;
begin
  if OptPagNum then begin
    Str(PageNum,PagNumStr);
    PagNumStr:= '-'+PagNumStr+'-';
  end
  else StpCreate(PagNumStr);
  if OptDblPrn and not odd(PageNum) then begin
    StpCpy(Header,PagNumStr);
    if not StpEmpty(Title) then begin
      StpcCat(Header,' ');
      StpFill(Header,' ',PagWid-StpLen(Title));
      StpCat(Header,Title);
    end;
  end
  else begin
    StpCpy(Header,Title);
    if OptPagNum then begin
      StpcCat(Header,' ');
      StpFill(Header,' ',PagWid-StpLen(PagNumStr));
      StpCat(Header,PagNumStr);
    end;
  end;
  if OptLinNum then Write(OutFvar,' ':LnumWid+2);
  WriteLn(OutFvar,Header); WriteLn(OutFvar);
end;


procedure PutPage;
var r: PageInd;
begin
  if OptPagNum or (not StpEmpty(Title)) then PutHeader;
  for r:= 1 to Nlines do begin
    inc(LineNum);
    if OptLinNum then Write(OutFvar,LineNum:LnumWid,': ');
    WriteLn(OutFvar,Page[r]);
  end;
  Write(OutFvar,chr(AsciiFF));
end;


procedure MainInit;
begin
  { set defaults: }
  Cols      := 1;
  PagLen    := 60;
  PagWid    := 80;
  Sep       := ' ';
  Title     := '';
  OptPagNum := false;
  OptLinNum := false;
  OptDblPrn := false;
  LnumWid   := 4;

  rc:= RcOk; InpOpen:= false; OutOpen:= false;
  ReadArgs;
  if rc = RcOk then begin
    SepWid:= StpLen(Sep);
    ColWid:= (PagWid - (Cols-1) * SepWid) div cols;
    OpenInp;
    if rc = RcOk then OpenOut;
  end;
end;


procedure MainTerm;
begin
  if InpOpen then Close(InpFvar);
  if OutOpen then Close(OutFvar);
  if rc <> RcOk then begin
    if rc=ErrArg then Help
    else begin
      write(ErrMsg[rc]);
      if      rc = ErrInp then write(InpFname)
      else if rc = ErrOut then write(OutFname);
      writeln;
    end;
  end;
end;


begin { Main program }
  MainInit;
  if rc = RcOk then begin
    PageNum:= 0; LineNum:= 0;
    while GetPage do begin
      inc(PageNum);
      PutPage;
    end;
    if OptDblPrn and odd(PageNum) then Write(OutFvar,chr(AsciiFF));
  end;
  MainTerm;
end.
